==> eh-0.4.tar.gz: eh-0.4/src/ehmemtrk.c <==
/***

ehmemtrk.c -- Error Handler, memory tracking (public interface)

Copyright (c) 1992-6 Christopher Cobb, All rights reserved.

***/

static const char*rcsid =
       "$Id: ehmemtrk.c,v 1.7 1996/03/08 07:35:10 ccobb Exp $";

/***

HACK ALERT: In order to support bounds checking, these functions
            treat the standard function free() as if it takes
            two parameters: the address of the memory to free,
            and the number of bytes of allocated memory.  This
            information, although not used at all by free(),
            is used by the bounds checking routines.  The length
            of the bounding block is a constant EH_SZBOUNDBLK.
***/

#include "eh.h"
#include "eherrs.h"

EHBEGINEXTERNC
#include <string.h>
EHENDEXTERNC

#include "ehbounds.h"
#include "ehintres.h"

#define EH_NMEMTRK 1
#include "ehmemtrk.h"

EHFILENAME(ehmemtrk.c);

/***
_ehMalloc()
Track dynamic memory allocation with malloc().
This function is a malloc() wrapper.  
***/

void *
_ehMalloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szBlk)
{
   void *volatile pvMem = NULL;
   volatile bool completed = false;

   EHCATCH(_ehMalloc) {

      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_MALLOC);
         break;
      }

      //printf("_ehMalloc: pvMem = 0x%p\n", pvMem);
      //printf("_ehMalloc: szBlk = %d\n",   szBlk);

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehMalloc: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehIntRes,
            (int)EH_RES_HEAPMEM, "malloc'ed memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHENDCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   EHRETURN(pvMem);
} /* _ehMalloc */


/***
_ehCalloc()
Track dynamic memory allocation with calloc().
This function is a calloc() wrapper.  
***/

void *
_ehCalloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szNum,
   size_t      szByt)
{
   void *volatile pvMem = NULL;
   size_t szBlk = szNum * szByt;
   volatile bool completed = false;

   EHCATCH(_ehCalloc) {

      // use malloc here in order to be able to add a bounds block
      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_CALLOC);
         break;
      }
      memset(pvMem, 0, szBlk);

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehCalloc: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehIntRes,
            (int)EH_RES_HEAPMEM, "calloc'ed memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHENDCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   // puts("_ehCalloc: leaving");

   EHRETURN(pvMem);
} /* _ehCalloc */


/***
_ehRealloc()
Track dynamic memory allocation with realloc().
This function is a realloc() wrapper.  
***/

void *
_ehRealloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   void       *pvMemOld,
   size_t      szBlk)
{
   void *volatile pvMemNew = pvMemOld;
   volatile bool completed = false;

   EHCATCH(_ehRealloc) {

      /* attempt to reallocate resource */
      pvMemNew = realloc(pvMemOld, szBlk + EH_SZBOUNDBLK);
      if (pvMemNew == NULL)
      {
         /* at this moment the old block is vunerable to being overwritten */
         ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_REALLOC); /* fatal error */
         break;
      }

      ehBndSet((char*)pvMemNew + szBlk, EH_SZBOUNDBLK);

      /* mark the resource as obtained */
      EHDBG1("_ehRealloc: pvMemNew = %lx\n", pvMemNew);
      completed = ehResReObt(pcFile,pcFunc,iLine,&ehIntRes,
            (int)EH_RES_HEAPMEM, "realloc'ed memory",
            (ulong)pvMemOld,
            (pfv_ulv)free, EH_RES_ALLOW_NULL | 2L, 
            (ulong)pvMemNew,
            (ulong)szBlk);

   } EHENDCATCH;

   if (ehError() && !completed)
   {
      if (pvMemNew && pvMemNew != pvMemOld) /* new block allocated */
      {
         /* old block *must* have been freed by realloc */
         free(pvMemNew); /* free the new block */
         pvMemNew = NULL;
         /* now both blocks are freed */
      }
      if (pvMemOld != NULL) /* realloc allows null */
      {
         /* mark the old block as freed */
         ehResRel(pcFile,pcFunc,iLine,&ehIntRes,
               (int)EH_RES_HEAPMEM, (pfv_ulv)0, 1L,
               (ulong)pvMemOld);
      }
   }
   EHRETURN(pvMemNew);
} /* _ehRealloc */


/***
_ehDynArr()
Track dynamic memory allocation of ehDynArr()
which manages extended dynamic arrays.

On first invocation, *pszNumEl == 0 and
*ppvArray is set to the initial array
location.  The resource handle will
have a specific address.

On subsequent invocations, the array
may stay in the same place or be moved.
In either case, the size of the array
increases.  The easiest way to deal
with this is to release the old resource
handle and obtain a new one with the
new address and size.

***/

#include "ehdynarr.h"

void *
_ehDynArr(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   void **ppvArray,
   size_t *pszNumEl,
   const size_t szEl)
{
   char * volatile pc = NULL;
   int   iNewNumEl;
   int   iNum2Alloc;
   char **ppcArray = (char **)ppvArray;

   EHCATCH(_ehDynArr) {

      EHASSERT_EXITSEQ((*ppcArray == NULL && *pszNumEl == 0) ||
             (*ppcArray != NULL && *pszNumEl != 0));
      EHASSERT_EXITSEQ(szEl > 0);

      iNewNumEl = (int)(*pszNumEl + 1);
      iNum2Alloc = (int)ehRoundUp(iNewNumEl);   /* # of elements to allocate */

      /* first time, prime the array */
      if (*ppcArray == NULL)
      {
         if ((*ppcArray = (char *)_ehMalloc(pcFile, pcFunc, iLine,
               (size_t)(szEl * iNum2Alloc)))
               == NULL)
         {
            ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_DYNARR);
            break;
         }
         pc = *ppcArray;
      }
      else /* extend dynamic array if necessary */
      {
         int iNumAllocated = ehRoundUp((uint)*pszNumEl);/*allocated last time*/

         if (iNum2Alloc > iNumAllocated) /* time to reallocate */
         {
            if ((*ppcArray = (char *)_ehRealloc(pcFile, pcFunc, iLine,
                  *ppcArray, (size_t)(szEl * iNum2Alloc)))
                  == NULL)
            {
               ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_DYNARR);
               break;
            }
         }
         pc = *ppcArray + (*pszNumEl * szEl);
      }
      EHDBG2("_ehDynArr: ppcArray = %lx, pc = %lx\n", ppcArray, pc);
      memset(pc, 0, (size_t)szEl);
      (*pszNumEl)++;

   } EHENDCATCH;

   if (ehError())
   {
      if (*ppcArray) _ehFree(pcFile, pcFunc, iLine, *ppcArray);
      *ppcArray = pc = NULL;
      *pszNumEl = 0;
      EHSTOPCHAIN();
   }
   EHRETURN( (void *) pc );
} /* _ehDynArr */


/***
_ehFree()
Track heap memory de-allocation.
***/

void
_ehFree(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   void       *pvMem)
{
   EHCATCH(_ehFree) {

   } EHENDCATCH;

   /* mark the resource and release it */
   EHDBG1("_ehFree: pvMem = %lx\n", pvMem);
   EHDBG1("_ehFree: pvMem = [%37.37s]\n", pvMem);
   ehResRel(pcFile,pcFunc,iLine,&ehIntRes,
         (int)EH_RES_HEAPMEM, (pfv_ulv)free, 1L, 
         (ulong)pvMem);

   EHLEAVE();
} /* _ehFree */


#ifdef ehEXCEPTIONHANDLING

/***
_ehNew()
Track dynamic memory allocation with malloc().
This function is a malloc() wrapper.  
***/

static
void *
_ehNew(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szBlk)
{
   void *volatile pvMem = NULL;
   volatile bool completed = false;

   EHCATCH(_ehNew) {

      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_NEW);
         break;
      }

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehNew: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehIntRes,
            (int)EH_RES_HEAPMEM, "new memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHENDCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   EHRETURN(pvMem);
} /* _ehMalloc */

/***
operator new wrapper.
***/

void *
operator new(
   size_t      szBlock,
   const char *pcFile,
   const char *pcFunc,
   int         iLine)
{
   EHTRACE(operator new);
   void *pv = _ehNew(pcFile,pcFunc,iLine,szBlock);
   EHDBG1("operator new: pv = %lx\n", pv);
   if (!pv)
      ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_NEW);
   return(pv);
}

#ifdef HAVE_ARRAYMEM

/***
operator new [] wrapper.
***/

void *
operator new[](
   size_t      szBlock,
   const char *pcFile,
   const char *pcFunc,
   int         iLine)
{
   EHTRACE(operator new[]);
   void *pv = _ehNew(pcFile,pcFunc,iLine,szBlock);
   EHDBG1("operator new[]: pv = %lx\n", pv);
   if (!pv)
      ehPError(pcFile, pcFunc, iLine, (schar)-1, EH_NEWARR);
   return(pv);
}

#endif /* HAVE_ARRAYMEM */

int         EHDELETE::de_iChan;
const char *EHDELETE::de_pcFile;
const char *EHDELETE::de_pcFunc;
int         EHDELETE::de_iLine;

EHDELETE::EHDELETE(
   int iChan,
   const char*pcFile,
   const char*pcFunc,
   int iLine)
{
   de_iChan  = iChan;
   de_pcFile = pcFile;
   de_pcFunc = pcFunc;
   de_iLine  = iLine;
}

/***
operator delete wrapper.
***/

void
operator delete(
   void       *pvMem)
{
   //EHTRACE(operator delete);
   if (EHDELETE::de_iLine) // this is one of our calls
   {
      int iLine = EHDELETE::de_iLine;
      EHDELETE::de_iLine = 0; // prevent recursive calls
      //fprintf(EHSTDERR, "operator delete: pvMem = %lx\n", pvMem);
      _ehFree(EHDELETE::de_pcFile, EHDELETE::de_pcFunc,iLine,pvMem);
   }
   else
   {
      free(pvMem);
   }
}

#ifdef HAVE_ARRAYMEM

/***
operator delete [] wrapper.
***/

void
operator delete[](
   void       *pvMem)
{
   //EHTRACE(operator delete[]);
   if (EHDELETE::de_iLine) // this is one of our calls
   {
      int iLine = EHDELETE::de_iLine;
      EHDELETE::de_iLine = 0; // prevent recursive calls
      EHDBG1("operator delete[]: pvMem = %lx\n", pvMem);
      _ehFree(EHDELETE::de_pcFile, EHDELETE::de_pcFunc,iLine,pvMem);
   }
   else
   {
      free(pvMem);
   }
}

#endif /* HAVE_ARRAYMEM */

#endif /* ehEXCEPTIONHANDLING */

==> eh-0.5.tar.gz: eh-0.5/src/ehmemtrk.c <==
/***

ehmemtrk.c -- Error Handler, memory tracking (public interface)

    Copyright (C) 1997  CSEG, Inc., All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                   

    The GNU General Public License is in the file COPYING which resides
    in the top directory of the distribution tree.

***/

static const char*rcsid =
       "$Id: ehmemtrk.c,v 1.7 1996/03/08 07:35:10 ccobb Exp $";

/***

HACK ALERT: In order to support bounds checking, these functions
            treat the standard function free() as if it takes
            two parameters: the address of the memory to free,
            and the number of bytes of allocated memory.  This
            information, although not used at all by free(),
            is used by the bounds checking routines.  The length
            of the bounding block is a constant EH_SZBOUNDBLK.
***/

#define EH_NO_MEMORY_TRACKING 1
#include "ehmemtrk.h"
#include "eherrs.h"

EHBEGINEXTERNC
#include <string.h>
EHENDEXTERNC

#include "ehbounds.h"
#include "ehintres.h"

EHFILENAME(ehmemtrk.c);

/***
_ehMalloc()
Track dynamic memory allocation with malloc().
This function is a malloc() wrapper.  
***/

void *
_ehMalloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szBlk)
{
   void *volatile pvMem = NULL;
   volatile bool completed = false;

   EHTRY(_ehMalloc) {

      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_MALLOC);
         break;
      }

      //printf("_ehMalloc: pvMem = 0x%p\n", pvMem);
      //printf("_ehMalloc: szBlk = %d\n",   szBlk);

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehMalloc: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehResTrk,
            (int)EH_RES_C_MEM, "malloc'ed memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   EHRETURN(pvMem);
} /* _ehMalloc */


/***
_ehCalloc()
Track dynamic memory allocation with calloc().
This function is a calloc() wrapper.  
***/

void *
_ehCalloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szNum,
   size_t      szByt)
{
   void *volatile pvMem = NULL;
   size_t szBlk = szNum * szByt;
   volatile bool completed = false;

   EHTRY(_ehCalloc) {

      // use malloc here in order to be able to add a bounds block
      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_CALLOC);
         break;
      }
      memset(pvMem, 0, szBlk);

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehCalloc: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehResTrk,
            (int)EH_RES_C_MEM, "calloc'ed memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   // puts("_ehCalloc: leaving");

   EHRETURN(pvMem);
} /* _ehCalloc */


/***
_ehRealloc()
Track dynamic memory allocation with realloc().
This function is a realloc() wrapper.  
***/

void *
_ehRealloc(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   void       *pvMemOld,
   size_t      szBlk)
{
   void *volatile pvMemNew = pvMemOld;

   EHTRY(_ehRealloc)
   {

      /* attempt to reallocate resource */
      pvMemNew = realloc(pvMemOld, szBlk + EH_SZBOUNDBLK);

   /***
      At this moment, the realloc block may have moved.  If it 
      did, the old block has been freed and is open to being 
      overwritten.  On some systems, the old bounding block 
      may already have been modified.
      
      Calling an eh-function (tracer/catcher) will cause bounds 
      checking to be performed on the old block, which the Error 
      Handler does not yet know has been freed.  Therefore, we want 
      to quickly mark the block as "released" before anything else 
      happens.
   ***/

      EH_BNDS_CHK_OFF();
      EH_CTRLC_CHK_OFF();

      if (pvMemNew != NULL) /* success */
      {

         /* !eh-function! */
         ehResReObt(pcFile,pcFunc,iLine,&ehResTrk,
               (int)EH_RES_C_MEM, "realloc'ed memory",
               (ulong)pvMemOld,
               (pfv_ulv)free, EH_RES_ALLOW_NULL | 2L, 
               (ulong)pvMemNew,
               (ulong)szBlk);

         /* non-eh-function */
         ehBndSet((char*)pvMemNew + szBlk, EH_SZBOUNDBLK);

      }
      else /* pvMemNew == NULL */
      {
         /* non-eh-function */
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_REALLOC);
         break;
      }

   } EHCATCH;

   if (ehError())
   {
      /* do this first before we touch any more memory */
      if (pvMemOld != NULL) /* realloc allows null */
      {
         /* !eh-function! mark the old block as freed */
         ehResRel(pcFile,pcFunc,iLine,&ehResTrk,
               (int)EH_RES_C_MEM, (pfv_ulv)0, 1L,
               (ulong)pvMemOld);
      }
      if (pvMemNew && pvMemNew != pvMemOld) /* new block allocated */
      {
         /* old block *must* have been freed by realloc */
         free(pvMemNew); /* free the new block */
         pvMemNew = NULL;
         /* now both blocks are freed */
      }
   }

   EH_CTRLC_CHK_ON();
   EH_BNDS_CHK_ON();
   EHRETURN(pvMemNew);
} /* _ehRealloc */


/***
_ehDynArr()
Track dynamic memory allocation of ehDynArr()
which manages extended dynamic arrays.

On first invocation, *pszNumEl == 0 and
*ppvArray is set to the initial array
location.  The resource handle will
have a specific address.

On subsequent invocations, the array
may stay in the same place or be moved.
In either case, the size of the array
increases.  The easiest way to deal
with this is to release the old resource
handle and obtain a new one with the
new address and size.

***/

#include "ehdynarr.h"

void *
_ehDynArr(
   const char *pcFile,
   const char *pcFunc,
   int iLine,
   void **ppvArray,
   size_t *pszNumEl,
   const size_t szEl)
{
   char * volatile pc = NULL;
   int   iNewNumEl;
   int   iNum2Alloc;
   char **ppcArray = (char **)ppvArray;

   EHTRY(_ehDynArr) {

      EHASSERT((*ppcArray == NULL && *pszNumEl == 0) ||
             (*ppcArray != NULL && *pszNumEl != 0));
      EHASSERT(szEl > 0);

      iNewNumEl = (int)(*pszNumEl + 1);

      /* # of elements to allocate */
      iNum2Alloc = (int)ehRoundUp(iNewNumEl);

      /* first time, prime the array */
      if (*ppcArray == NULL)
      {
         if ((*ppcArray = (char *)_ehMalloc(pcFile, pcFunc, iLine,
               (size_t)(szEl * iNum2Alloc)))
               == NULL)
         {
            ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_DYNARR);
            break;
         }
         pc = *ppcArray;
      }
      else /* extend dynamic array if necessary */
      {
         /*allocated last time*/
         int iNumAllocated = ehRoundUp((uint)*pszNumEl);

         if (iNum2Alloc > iNumAllocated) /* time to reallocate */
         {
            if ((*ppcArray = (char *)_ehRealloc(pcFile, pcFunc, iLine,
                  *ppcArray, (size_t)(szEl * iNum2Alloc)))
                  == NULL)
            {
               ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_DYNARR);
               break;
            }
         }
         pc = *ppcArray + (*pszNumEl * szEl);
      }
      EHDBG2("_ehDynArr: ppcArray = %lx, pc = %lx\n", ppcArray, pc);
      memset(pc, 0, (size_t)szEl);
      (*pszNumEl)++;

   } EHCATCH;

   if (ehError())
   {
      if (*ppcArray) _ehFree(pcFile, pcFunc, iLine, *ppcArray);
      *ppcArray = pc = NULL;
      *pszNumEl = 0;
      //EHSTOPCHAIN();
   }
   EHRETURN( (void *) pc );
} /* _ehDynArr */


/***
_ehFree()
Track heap memory de-allocation.
***/

void
_ehFree(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   void       *pvMem)
{
   EHTRY(_ehFree) {

   } EHCATCH;

   /* mark the resource and release it */
   EHDBG1("_ehFree: pvMem = %lx", pvMem);
   EHDBG1("_ehFree: pvMem = [%37.37s]", pvMem);
   ehResRel(pcFile,pcFunc,iLine,&ehResTrk,
         (int)EH_RES_C_MEM, (pfv_ulv)free, 1L, 
         (ulong)pvMem);

   EHLEAVE();
} /* _ehFree */


#ifdef ehEXCEPTIONHANDLING

/***
_ehNew()
Track dynamic memory allocation with malloc().
This function is a malloc() wrapper.  
***/

static
void *
_ehNew(
   const char *pcFile,
   const char *pcFunc,
   int         iLine,
   size_t      szBlk)
{
   void *volatile pvMem = NULL;
   volatile bool completed = false;

   EHTRY(_ehNew) {

      pvMem = malloc(szBlk + EH_SZBOUNDBLK);
      if (pvMem == NULL)
      {
         ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_NEW);
         break;
      }

      ehBndSet((char*)pvMem + szBlk, EH_SZBOUNDBLK);

      EHDBG1("_ehNew: pvMem = %lx\n", pvMem);
      completed = ehResObt(pcFile,pcFunc,iLine,&ehResTrk,
            (int)EH_RES_CPP_MEM, "new memory",
            (pfv_ulv)free, 2L, 
            (ulong)pvMem,
            (ulong)szBlk);

   } EHCATCH;

   if (ehError() && !completed)
   {
      if (pvMem)
      {
         free(pvMem);
         pvMem = NULL;
      }
   }

   EHRETURN(pvMem);
} /* _ehMalloc */

/***
operator new wrapper.
***/

void *
operator new(
   size_t      szBlock,
   const char *pcFile,
   const char *pcFunc,
   int         iLine)
{
   EHTRACE(operator new);
   void *pv = _ehNew(pcFile,pcFunc,iLine,szBlock);
   EHDBG1("operator new: pv = %lx\n", pv);
   if (!pv)
      ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_NEW);
   return(pv);
}

#ifdef HAVE_ARRAYMEM

/***
operator new [] wrapper.
***/

void *
operator new[](
   size_t      szBlock,
   const char *pcFile,
   const char *pcFunc,
   int         iLine)
{
   EHTRACE(operator new[]);
   void *pv = _ehNew(pcFile,pcFunc,iLine,szBlock);
   EHDBG1("operator new[]: pv = %lx\n", pv);
   if (!pv)
      ehPError(pcFile, pcFunc, iLine, EH_IS_FATAL, EH_NEWARR);
   return(pv);
}

#endif /* HAVE_ARRAYMEM */

int         EHDELETE::de_iChan;
const char *EHDELETE::de_pcFile;
const char *EHDELETE::de_pcFunc;
int         EHDELETE::de_iLine;

EHDELETE::EHDELETE(
   int iChan,
   const char*pcFile,
   const char*pcFunc,
   int iLine)
{
   de_iChan  = iChan;
   de_pcFile = pcFile;
   de_pcFunc = pcFunc;
   de_iLine  = iLine;
}

/***
operator delete wrapper.
***/

void
operator delete(
   void       *pvMem)
{
   if (EHDELETE::de_iLine) // this is one of our calls
   {
      int iLine = EHDELETE::de_iLine;
      EHDELETE::de_iLine = 0; // prevent recursive calls
      ehResRel(EHDELETE::de_pcFile, EHDELETE::de_pcFunc,iLine,&ehResTrk,
            (int)EH_RES_CPP_MEM, (pfv_ulv)free, 1L, 
            (ulong)pvMem);
   }
   else
   {
      free(pvMem);
   }
}

#ifdef HAVE_ARRAYMEM

/***
operator delete [] wrapper.
***/

void
operator delete[](
   void       *pvMem)
{
   if (EHDELETE::de_iLine) // this is one of our calls
   {
      int iLine = EHDELETE::de_iLine;
      EHDELETE::de_iLine = 0; // prevent recursive calls
      ehResRel(EHDELETE::de_pcFile, EHDELETE::de_pcFunc,iLine,&ehResTrk,
            (int)EH_RES_CPP_MEM, (pfv_ulv)free, 1L, 
            (ulong)pvMem);
   }
   else
   {
      free(pvMem);
   }
}

#endif /* HAVE_ARRAYMEM */

#endif /* ehEXCEPTIONHANDLING */
